/* 2012.Feb create functions of Markov Calculations */


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "snp_Config.h"
#include "snp_Random.h"
#include "snp_Markov.h"

dsfmt_t snp_dsfmt_global_data;



int SetMType(char *MarkovType){
	int mtype=0;
	if(strcmp(MarkovType,"Original")==0){
		mtype=1;
	}else if(strcmp(MarkovType,"Fast")==0){
		mtype=2;
	}else if(strcmp(MarkovType,"FastDouble")==0){
		mtype=10;
	}else if(strcmp(MarkovType,"Fast2")==0){
		mtype=3;
	}else if(strcmp(MarkovType,"Rat")==0){
		mtype=0;
	}

	return mtype;
}

void DoMarkovDouble(int M_Type,double **Xdouble,double *h,long L,dsfmt_t *snp_dsfmt_data){
	switch(M_Type){
		case(10):{
			TypeIMarkovFastDouble(Xdouble,h,L,snp_dsfmt_data);
			break;
		}
	}
}


void DoMarkov(int M_Type,int **X,double *h,long L,dsfmt_t *snp_dsfmt_data){
	switch(M_Type){
		case(1):{
			TypeIMarkov(X,h,L,snp_dsfmt_data);
			break;
			}
		case(2):{
			TypeIMarkovFast(X,h,L,snp_dsfmt_data);
			break;
			}
		case(3):{
			TypeIMarkov2(X,h,L,snp_dsfmt_data);
			break;
			}
	}
}




void TypeIMarkov(int **X, double *h,long L,dsfmt_t *snp_dsfmt_data)
{
	int j=0;
	long u=0;
	long v=0;
	int XjuStar=0;
	int XjvStar=0;

	double c=0;
	double tmp=0;

	/* procedure 3 */
	j=TypeIZeroOne(snp_dsfmt_data);

	/* procedure 4 */
	u=(int)(myRand(snp_dsfmt_data)*L);

	if(0==X[j][u]){		/* condition 5 */
		/* invariant */
	} else {		/* condition 6 */
		v=0;
		do {
			v=(int)(myRand(snp_dsfmt_data)*L);
		} while(u==v);

		/* new candidate */
		XjuStar=X[j][u]-1;
		XjvStar=X[j][v]+1;

		/* transition probability */
		c=h[v]*X[j][u]/(h[u]*(X[j][v]+1));

		if(c>=1){	/* condition 8 */
			X[j][u]=XjuStar;
			X[j][v]=XjvStar;
		} else {
			tmp=myRand(snp_dsfmt_data);
			if(tmp<c){
				X[j][u]=XjuStar;
				X[j][v]=XjvStar;
			}
		}
		/* procedure 10 */
	}
}





void TypeIMarkovFast(int **X,double *h,long L,dsfmt_t *snp_dsfmt_data){
	int target=0;

	target=(int)myRand(snp_dsfmt_data)*(2*L);	/* choose one of haplotypes */

	/* procedure 3 */
	int j=target/L;		/* 0 or 1 */

	/* procedure 4 */
	int /* long */ u=target/2;		/* 0,.....,L-1 */
	int /* long */ v=0;

	if(0<X[j][u]){		/* condition 5 */ /* condition 6 */
		v=0;
		// do {
		// v=(int)(myRandFast()*L);
		// } while(u==v);

		v=(int)(myRand(snp_dsfmt_data)*(L-1));
		if(v>=u) v++;
		
		/* transition probability */
		if(h[v]*X[j][u]>=(h[u]*(X[j][v]+1))){	/* condition 8 */
			/* new candidate */
			X[j][u]--;
			X[j][v]++;
		} else {
			if(myRand(snp_dsfmt_data)*(h[u]*(X[j][v]+1))<h[v]*X[j][u]){
				/* new candidate */
				X[j][u]--;
				X[j][v]++;
			}
		}
		/* procedure 10 */
	}
}

void TypeIMarkovFastDouble(double **X,double *h,long L,dsfmt_t *snp_dsfmt_data){
	int target=0;
	target=(int)(myRand(snp_dsfmt_data)*(2*L));	/* choose one of haplotypes */

	/* procedure 3 */
	int j=target/L;	/* 0 or 1 */
	
	/* procedure 4 */
	int /* long */ u=target/2;	/* 0,....,L-1 */
	int /* long */ v=0;

	if(0.5<X[j][u]){	/* condition 5 */ /* condirion 6 */
		// v=0;
		// do {
		// v=(int)(myRandFast()*L);
		// } while(u==v);

		v=(int)(myRand(snp_dsfmt_data)*(L-1));
		if(v>=u) v++;

		/* transition posibility */
		if(h[v]*X[j][u]>=(h[u]*X[j][v]+1)){	/* condition 8 */
			/* new candidate */
			X[j][u]--;
			X[j][v]++;
		} else {
			if(myRand(snp_dsfmt_data)*(h[u]*(X[j][v]+1))<h[v]*X[j][u]){
				/* new candidate */
				X[j][u]--;
				X[j][v]++;
			}
		}
		/* procedure 10 */
	}
}


void TypeIMarkov2(int **X,double *h,long L, dsfmt_t *snp_dsfmt_data){
	double tmp=myRand(snp_dsfmt_data)*2*L*(L-1);
	int v=(int)(tmp);
	tmp-=v;			/* 0<=tmp<1 */
	int j=(v&1);		/* 0 or 1 */
	v>>=1;
	int u=(v%L);		/* 0,...,L-1 */
	v/=L;			/* 0,...,L-1 */
	if(v>=u) v++;
	int Xjv=X[j][v];
	int Xju=X[j][u];
	if(0<Xju){
		if (tmp*(h[u]*(Xjv+1))<h[v]*Xju){
			Xju--;
			Xjv++;
		}
		X[j][v]=Xjv;
		X[j][u]=Xju;
	}
}



int TypeIZeroOne(dsfmt_t *snp_dsfmt_data)
{
 	/* return 0 or 1 with probability 0.5 */
 	int result=0;
 	result=(int)myRand(snp_dsfmt_data)*2;
 	return result; 
}


