#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <mpi.h>

#ifdef _OPENMP
#include <omp.h>
#endif

#define _MAIN_DEF


#ifdef MPI_USE
/* プロトタイプ宣言 */
int Define_MPI_InputStruct(MPI_Datatype *mpi_inputStruct);
#endif

#include "snp_RATParallel.h"

dsfmt_t snp_dsfmt_global_data;
#pragma omp threadprivate(snp_dsfmt_global_data)



/* RATのPermutation検定を実行する */
int RATExecute(SnpData *snpData1, SnpData *snpData2, double Sobs, long a, long b, long jStart, long jEnd, double *typeIError, long repeat, long gen,int flag)
{
    long h = 0;
    long i = 0;
    long j = 0;
    long m = 0;
    long n = 0;
    int burnin = 100000; //10万に変更

    int threadNum = 1;      /* スレッド数 */
    int iam = 0;            /* スレッド番号 */
#ifdef _OPENMP
    /* 使用可能なプロセッサ数を取得 */
    threadNum = omp_get_num_procs();
#endif


    n = a + b;
    m = jEnd - jStart;

    /* significantではない場合、終了 */
    if ( Sobs < 3.84){
//        printf("P > 0.05\n");
        return 0.05;
    }

    /* log テーブルを作成する */
    FactorialSetFactorial(n);



// OpenMP スレッド並列開始
#pragma omp parallel private(h, iam)
    {
#ifdef _OPENMP
        /* 自分のスレッド番号を取得 */
        iam = omp_get_thread_num();
#endif
//        int repeat = threadNum;//将来は変更
        int r=0;
        double S = 0;
        long gen2 = 0;
#pragma omp for private(j, flag, gen2, S)
        for (r = 0; r < repeat; r++){

            double sumOfQj = 0;
            double sumOfinvQj = 0;
            double Srat = 0;
            double F = 0;
            double logF = 0;
            double Qvalue = 0;
            SnpTable T;
            int *type = NULL;
            int *di = NULL;
            int **genotype = NULL;
            int ***Tj = NULL;
            double *Sj = NULL;
            double *numberOfDiInHj = NULL;


        	/* typeのメモリ確保 */
            type = (int*)malloc1Dim(sizeof(int), n);
            if (NULL == type) { goto finalize; }
            /* diのメモリ確保 */
            di = (int*)malloc1Dim(sizeof(int), n);
            if (NULL == di) { goto finalize; }
            /* genotypeのメモリ確保 */
            genotype = (int**)mallocInt2Dim(m, n);
            if (NULL == genotype) { goto finalize; }
            /* Tjのメモリ確保 */
            Tj = (int***)mallocInt3Dim(m, ROW, COLUMN);
            if (NULL == Tj) { goto finalize; }
            /* Sjのメモリ確保 */
            Sj = (double*)malloc1Dim(sizeof(double), m);
            if (NULL == Sj) { goto finalize; }
            /* numberOfDiInHjのメモリ確保 */
            numberOfDiInHj = (double*)malloc1Dim(sizeof(double), m);
            if (NULL == numberOfDiInHj) { goto finalize; }

            for (i=0,j = jStart; j < jEnd; j++){
                /* 観測値から偶現表を作成する */
                DataReaderPopulationType(&snpData1[j], &snpData2[j], type);
                DataReaderMakeTableDi(&snpData1[j], &snpData2[j], type, Tj[i]);
                /* genotypeを計算 */
                RATGenotype(&snpData1[j], &snpData2[j], genotype[i]);
                i++;
            }

            for (i = 0; i < m; i++){
                TableMakeTableIntArray(&T, Tj[i]);
                numberOfDiInHj[i] = RATNumberOfDiInHj(T, Sobs);
                sumOfQj += numberOfDiInHj[i];
                /* 構造体tableのメモリ開放 */
                TableFinalTable(&T);
            }
            
            for (h= 0; h < burnin; h++){
                for (i = 0; i < m; i++){
                    TableMakeTableIntArray(&T, Tj[i]);
                    RATAlgorithmB(T, Sobs, Tj[i], &snp_dsfmt_global_data);
                    /* 構造体tableのメモリ開放 */
                    TableFinalTable(&T);
                }
            }


            for (h = 0; h < gen; h++){
                j = RATAlgorithmA(numberOfDiInHj, m, &snp_dsfmt_global_data); 
                TableMakeTableIntArray(&T, Tj[j]);
                RATAlgorithmB(T, S, Tj[j], &snp_dsfmt_global_data);
                RATAlgorithmC(genotype[j], Tj[j], di, &snp_dsfmt_global_data);
                Qvalue = RATCalcQvalue(genotype, di, Sobs, n, m);
                if (Qvalue != 0){
                    sumOfinvQj += 1 / Qvalue;
                    gen2++;
                }
                /* 構造体tableのメモリ開放 */
                TableFinalTable(&T);
            }
            //F = FactorialGetCombination(n, a);
            logF = FactorialGetLogCombination(n, a);
            if (0 == gen2){
                typeIError[r] = -1;
            }
            else{
                typeIError[r] = exp( log(sumOfQj) + log(sumOfinvQj) - log(gen2) - logF );
            //result = sumOfQj * sumOfinvQj / gen2 / F;
            }
finalize:;
             /* 確保したメモリを開放する */
             free1Dim(type);
             freeInt2Dim(genotype, m);
             freeInt3Dim(Tj, m, ROW);
             free1Dim(Sj);
             free1Dim(numberOfDiInHj);
             free1Dim(di);
             /* 構造体tableのメモリ開放 */
             TableFinalTable(&T);
         }
         /* logテーブルクリア */
         FactorialDeleteFactorial();
        // end of repeat
    }
    return 0;
}

/* マーカーjをnumberOfDiInHj[j]に比例した確率で返す */
long RATAlgorithmA(double *numberOfDiInHj, long m, dsfmt_t *snp_dsfmt_data)
{
    long j = 0;
    double p = 0;
    double sumOfNumberOfHj = 0;


    for (j = 0; j < m; j++){
        sumOfNumberOfHj += numberOfDiInHj[j];
    }
    p = sumOfNumberOfHj * myRand(snp_dsfmt_data);  

    for (j = 0; j < m; j++){
        if (p < numberOfDiInHj[j]){
            return j;
        }
        p -= numberOfDiInHj[j];
    }

    return 0;
}

/* 超幾何分布に従うMCMCで、以前のテープルから次のテープルを作成 */
int RATAlgorithmB(SnpTable T, double Score, int **table, dsfmt_t *snp_dsfmt_data)
{
    if (FLAG_TRUE == TableTwoParts(T, Score)){
        TableMarkov2(T, Score, table, snp_dsfmt_data);
    }
    else{
        TableMarkov1(T, Score, table, snp_dsfmt_data);
    }

    return 0;
}

/* 偶現表から１つのpermutation eventを等確率で抽出 */
int RATAlgorithmC(int *genotype, int **table, int *di, dsfmt_t *snp_dsfmt_data)
{
    PermutationConditionedDi(genotype, table, di, snp_dsfmt_data);
    return 0;
}

/* Qの値を計算 */
long RATCalcQvalue(int **genotype, int *di, double Sobs, long n, long dataNum)
{
    int retval = 0;
    long j = 0;
    long result = 0;
    double S = 0;
    int **T = NULL;

    /* Tのメモリ確保 */
    T = (int**)mallocInt2Dim(ROW, COLUMN);
    if (NULL == T){ goto finalize; }

    for (j = 0; j < dataNum; j++){
        retval = PermutationCalcT(di, genotype[j], T, n);
        S = TableCalcScore(T);
        if (S >= Sobs){
            result++;
        }      	
    }

finalize:;
    /* 確保したメモリを開放する */
    freeInt2Dim(T, ROW);

    return result;
}

double RATNumberOfDiInHj(SnpTable table, double Sobs)
{
    double result = 0;
    int l = 0;
    int u = 0;
    int retval = 0;
    int **T = NULL;

    int i = 0;
    int a1 = 0;

    /* Tのメモリ確保 */
    T = (int**)mallocInt2Dim(ROW, COLUMN);
    if (NULL == T) {
        retval = 1;
        goto finalize;
    }

    l = TableCalcL(table);
    u = TableCalcU(table);

    for (a1 = l; a1 <= u; a1++){
        retval = TableNewTable(table, a1, T);
        if (FLAG_TRUE == TableInCj(Sobs, T, ROW, COLUMN)){
            result += TableCalcMu(T);
        }
    }

finalize:;
    /* 確保したメモリを開放する */
    freeInt2Dim(T, ROW);

    return result;
}

/* genotypeを計算 */
int RATGenotype(SnpData *snpData1, SnpData *snpData2, int *genotype)
{
    long h = 0;
    long a = 0;
    long b = 0;
    char reference = 0;

    int *genotype1 = NULL;
    int *genotype2 = NULL;

    a = snpData1->dataNum;
    b = snpData2->dataNum;
   
    /* genotype1のメモリ確保 */
    genotype1 = (int*)malloc1Dim(sizeof(int), a);
    if (NULL == genotype1) { goto finalize; }
    /* genotype2のメモリ確保 */
    genotype2 = (int*)malloc1Dim(sizeof(int), b);
    if (NULL == genotype2) { goto finalize; }

    /* 参照文字を決定 */
    reference = DataReaderMaximumNuc(snpData1);

    /* genotypeを計算 */
    DataReaderCalcGenotype(snpData1, reference, genotype1);
    DataReaderCalcGenotype(snpData2, reference, genotype2);
    for (h = 0; h < a; h++){
        genotype[h] = genotype1[h];
    }
    for (h = 0; h < b; h++){
        genotype[h + a] = genotype2[h];
    }

finalize:;
    /* 確保したメモリを開放する */
    free1Dim(genotype1);
    free1Dim(genotype2);

    return 0;
}

