/*!
  @file      snp_typeI_D.c
  @brief     typeIParallelD.exe Main program file
  @attention based on snp_typeI.c
  @date      March 2011
*/

/* modified 2012.Feb 						*/
/* add the function of selectable Markov Calculation		*/
/*	- Original 	original version Markov calc.		*/
/*	- FastRandom	changing generate random number 	*/
/*	- Fast		changing loops of generate random number */
/*	- FastDouble	using double type Markov calc.		*/
/* 	- Fast2							*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef MPI_USE
#include <mpi.h>
#endif

#ifdef _OPENMP
#include <omp.h>
#endif


#define _MAIN_DEF
#include "snp_typeI.h"
#include "SNP_DivFileXML.h"
#include "SNP_DivFile.h"

/* modified 2012.Feb add include file doe selectable Markov calc. */
#include "snp_typeI_Dfast.h"
dsfmt_t snp_dsfmt_global_data;

#ifdef MPI_USE
int Define_MPI_InputStruct(MPI_Datatype *mpi_inputStruct);
#endif


/* プロトタイプ宣言 */
void MainProgramMultiD(InputTypeI *inputTypeI, SNP_DivFileSet *divFileSet);


/** MPIの初期化処理
 *  MPI用の引数が処理されて、削除される。
 *  @param argc [IN/OUT] 引数の数
 *  @param argv [IN/OUT] 引数
 */
void setupMPI(int *argc, char ***argv)
{
#ifdef MPI_USE
    MPI_Init(argc, argv);
    MPI_Comm_size(MPI_COMM_WORLD, &MyMpiSize);
    MPI_Comm_rank(MPI_COMM_WORLD, &MyMpiRank);
#else
    MyMpiSize=1;
    MyMpiRank=0;
#endif		/* MPI_USE */
}


int main(int argc, char* argv[])
{
	/* modified 2012.Feb change data_structure of InputTypeI */

    InputTypeI inputTypeI={"","","",0,"",0,0,0,0,1};
//    InputTypeI inputTypeI={"", "", "", 0, "", 0, 0, 0, 0,""};	/* 2012.Feb for text parameter of MarkovType */
    char *inputFileXml;
    SNP_DivFileSet divFileSet;
    int i = 0;

    /* MPIによる初期化処理 */
    setupMPI(&argc, &argv);

#ifdef MPI_USE
    /* 時間計測用変数の初期化 */
    for (i = 0; i < NUM_TM; i++){
        tm[i] = 0.0;
    }

    /* 時間計測 */
    tm[TM_START] = MPI_Wtime();
#endif			/* MPI_USE */

    if(argc != 7){
        printf("[usage]typeIParallelDFast.exe [InputFile.xml]"
	       " [Score] [Repeat] [Generation] [OutputFile] [CalcType] \n");
        return 255;
    }
    inputFileXml = argv[1];
    inputTypeI.score = atoi(argv[2]);
    inputTypeI.repeat = atol(argv[3]);
    inputTypeI.gen = atol(argv[4]);
    strcpy(inputTypeI.outputFile1, argv[5]);

    /* modified 2012.Feb setting Markov calculation type */
	inputTypeI.CalcType=atoi(argv[6]);
//    strcpy(inputTypeI.MarkovType,argv[6]);

    /* XMLファイル入力  */
    Init_SNP_DivFileSet(&divFileSet);
    if ( ParseXMLFile(inputFileXml, &divFileSet) != 0 ) {
        return 255;
    }
    inputTypeI.dataType = divFileSet.caseDataType;
    inputTypeI.areaFileType = divFileSet.blockAreaType;
    strcpy(inputTypeI.inputFile1, divFileSet.caseFileName);
    strcpy(inputTypeI.inputFile2, divFileSet.controlFileName);
    strcpy(inputTypeI.blockAreaFile, divFileSet.blockAreaFileName);

    /* 実行条件表示 */
    if (MyMpiRank == 0) {
        printf("inputFile1: %s\n", inputTypeI.inputFile1);
        printf("inputFile2: %s\n", inputTypeI.inputFile2);
        printf("outputFile1: %s\n", inputTypeI.outputFile1);
        printf("blockAreaFile: %s\n", inputTypeI.blockAreaFile);
        printf("score: %d\n", inputTypeI.score);
        printf("repeat: %ld\n", inputTypeI.repeat);
        printf("gen: %ld\n", inputTypeI.gen);
        printf("dataType: %d\n", inputTypeI.dataType);
	/* modified 2012.Feb add printout of Markov calculation type */

	if(inputTypeI.CalcType==0){
		printf("RAT Calculation\n");
	}else{
		printf("Markov Calculation\n");
	}

//	if(strcmp(inputTypeI.MarkovType,"Rat")==0){
//		printf("Rat Calculation:\n");
//	}else{
//		printf("MarkovType: %s\n",inputTypeI.MarkovType);
//	}

        printf("MyMpiSize: %d\n", MyMpiSize);
    }

    /* スコア計算方法の指定 */
    iWay = inputTypeI.score;
#ifdef _OPENMP
#pragma omp parallel
#endif
{
    /* 乱数発生関数初期化 */
    initMyRand(&snp_dsfmt_global_data);
}

    /* MultiLocus検定を行う */

    MainProgramMultiD(&inputTypeI, &divFileSet);
    Clear_SNP_DivFileSet(&divFileSet);

#ifdef MPI_USE

    /* 時間計測 */
    tm[TM_END] = MPI_Wtime();

    MPI_Finalize();    /* MPI終了処理 */

	printf("%4d\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\t%lf\n",
		MyMpiRank,
		tm[TM_B_F_BCAST]-tm[TM_START],
		tm[TM_B_BCAST]  -tm[TM_B_F_BCAST],
		tm[TM_B_CALC]   -tm[TM_B_BCAST],
		tm[TM_A_CALC]   -tm[TM_B_CALC],
		tm[TM_A_GATHER] -tm[TM_A_CALC],
		tm[TM_END]      -tm[TM_A_GATHER],
		tm[TM_END]      -tm[TM_START]);
#if 0

    printf("%4d  1 time Init-READ      %lf (sec)\n", MyMpiRank, tm[TM_B_READ]     - tm[TM_START]);
    printf("%4d  2 time READ           %lf (sec)\n", MyMpiRank, tm[TM_A_READ]     - tm[TM_B_READ]);
    printf("%4d  3 time READ-FBcast    %lf (sec)\n", MyMpiRank, tm[TM_B_F_BCAST]  - tm[TM_A_READ]);
    printf("%4d  4 time FBcast         %lf (sec)\n", MyMpiRank, tm[TM_A_F_BCAST]  - tm[TM_B_F_BCAST]);
    printf("%4d  5 time FBcast-Bcast   %lf (sec)\n", MyMpiRank, tm[TM_B_BCAST]    - tm[TM_A_F_BCAST]);
    printf("%4d  6 time Bcast          %lf (sec)\n", MyMpiRank, tm[TM_A_BCAST]    - tm[TM_B_BCAST]);
    printf("%4d  7 time Bcast-Calc     %lf (sec)\n", MyMpiRank, tm[TM_B_CALC]     - tm[TM_A_BCAST]);
    printf("%4d  8 time Calculation    %lf (sec)\n", MyMpiRank, tm[TM_A_CALC]     - tm[TM_B_CALC]);
    printf("%4d  9 time Calc-FGather   %lf (sec)\n", MyMpiRank, tm[TM_B_F_GATHER] - tm[TM_A_CALC]);
    printf("%4d 10 time FGather        %lf (sec)\n", MyMpiRank, tm[TM_A_F_GATHER] - tm[TM_B_F_GATHER]);
    printf("%4d 11 time FGather-Gather %lf (sec)\n", MyMpiRank, tm[TM_B_GATHER]   - tm[TM_A_F_GATHER]);
    printf("%4d 12 time Gather         %lf (sec)\n", MyMpiRank, tm[TM_A_GATHER]   - tm[TM_B_GATHER]);
    printf("%4d 13 time Gather-Write   %lf (sec)\n", MyMpiRank, tm[TM_B_WRITE]    - tm[TM_A_GATHER]);
    printf("%4d 14 time Write          %lf (sec)\n", MyMpiRank, tm[TM_A_WRITE]    - tm[TM_B_WRITE]);
    printf("%4d 15 time Write-End      %lf (sec)\n", MyMpiRank, tm[TM_END]        - tm[TM_A_WRITE]);

    printf("%4d 20 time Init           %lf (sec)\n", MyMpiRank, tm[TM_B_F_BCAST]  - tm[TM_START]);
    printf("%4d 21 time FBcast         %lf (sec)\n", MyMpiRank, tm[TM_B_BCAST]    - tm[TM_B_F_BCAST]);
    printf("%4d 22 time Bcast          %lf (sec)\n", MyMpiRank, tm[TM_B_CALC]     - tm[TM_B_BCAST]);
    printf("%4d 23 time Calculation    %lf (sec)\n", MyMpiRank, tm[TM_A_CALC]     - tm[TM_B_CALC]);
    printf("%4d 24 time Gather         %lf (sec)\n", MyMpiRank, tm[TM_A_GATHER]   - tm[TM_A_CALC]);
    printf("%4d 25 time End            %lf (sec)\n", MyMpiRank, tm[TM_END]        - tm[TM_A_GATHER]);

    printf("%4d 30 time Calculation    %lf (sec)\n", MyMpiRank, tm[TM_A_CALC]     - tm[TM_B_CALC]);
    printf("%4d 31 time F_Bcast-Gather %lf (sec)\n", MyMpiRank, tm[TM_A_GATHER]   - tm[TM_B_F_BCAST]);
    printf("%4d 32 time A_Read-B_Write %lf (sec)\n", MyMpiRank, tm[TM_A_WRITE]    - tm[TM_B_READ]);
    printf("%4d 33 time Total          %lf (sec)\n", MyMpiRank, tm[TM_END]        - tm[TM_START]);

#endif 
#endif		/*MPI_USE */

    return 0;
}


/*! MultiLocus検定を行う
  @param   inputTypeI   入力データ格納用構造体へのポインタ
  @param   divFileSet   データ分割情報構造体へのポインタ
 */
void MainProgramMultiD(InputTypeI *inputTypeI, SNP_DivFileSet *divFileSet)
{
    int retval = 0;
    int flag = 0;
    int numOutput = 0;
    int outDataOffset = 0;
    int numPartOutput = 0;
    int numTIEPartOutput = 0;
    long i = 0;
    long j = 0;
    long k = 0;
    long l = 0;
    long areaFileLine = 0; /* haplotypeブロック領域指定ファイルのライン数 */
    long a = 0;         /* number of haplotype copies (=sequences) in case */
    long b = 0;         /* number of haplotype copies (=sequences) in control */
    long n = 0;
    long dataNum = 0;   /* case、controlデータ整合後の(このRankで担当する)SNP数 */
    int numData;
    int *numDataAll = NULL;/* case、controlデータ整合後の総SNP数 */
    long numDataTotal = 0;
    long index = 0;
    long jStart = 0;    /* haplotypeブロックの最初のSNPを示す座位 */
    long jEnd = 0;      /* haplotypeブロックの最後のSNPを示す座位 */
    long blockNumAll = 0;/* haplotype総ブロック数(全Rankの合計) */
    long repeat = 0;
    long startPos = 0;
    long endPos = 0;
    long blockPartNum = 0;
    long blockPartStart = 0;
    int start_FileIdx;
    long start_blockIdx;
    int end_fileIdx;
    long end_blockIdx;
    double S = 0;

    /* Gather用 */
    int *numOutputs = NULL;
    int *outDataOffsets = NULL;
    int *outTIEOutputs = NULL;
    int *outTIEDataOffsets = NULL;
    int *numTIEOutputs = NULL;
    int **T = NULL;                 /* 偶現表 */
    int *populationType = NULL;
    char ***haplotype = NULL;       /* haplotypeデータ */
    long *L = NULL;                 /* 各haplotypeブロック毎のhaplotypeパターン数 */
    long *blockArea = NULL;         /* 各haplotypeブロック領域格納 */
    long *blockAreaAll = NULL;      /* 全haplotypeブロック領域格納 */
    long *linkSNPNum = NULL;        /* 各haplotypeブロックのSNP数 */
    long *linkSNPStart = NULL;      /* 各haplotypeブロックの最初のSNPを示す座位 */
    long *maxScoreIndex = NULL;     /* 各haplotypeブロック内のスコア最大値の座位 */
    long *maxScoreIndexAll = NULL;  /* Haplotypeブロック内のスコア最大値の座位 */
    double **freq = NULL;           /* 頻度 */
    double *typeIError = NULL;      /* type I error 格納配列 */
    double *typeIErrorAll = NULL;   /* type I error 格納配列 */
    double *Sobs = NULL;            /* 各haplotypeブロック毎のスコア最大値 */
    double *SobsAll = NULL;         /* Haplotypeブロック毎のスコア最大値 */

    long *linkSNPNumAll = NULL;     /* 各haplotypeブロックの総SNP数 */
    long *linkSNPStartAll = NULL;   /* 各haplotypeブロックの最初のSNPを示す全体での座位 */
    char *snpDataAll_rsN = NULL;
    long *snpDataAll_pos = NULL;
    char *snpData_rsN = NULL;
    long *snpData_pos = NULL;
    int *outSnpRsNNums = NULL;
    int *outSnpPosNums = NULL;
    int *outSnpRsNOffsets = NULL;
    int *outSnpPosOffsets = NULL;
    long countData;

    FILE *fpOut = NULL;             /* 出力ファイルポインタ */

    SnpData *snpData1 = NULL;
    SnpData *snpData2 = NULL;

    int M_Type=1;		/* parameter of Markov calculation type */

    if ( ! inputTypeI || ! divFileSet ) {
      goto finalize;
    }

    repeat = inputTypeI->repeat;
    blockNumAll = divFileSet->blockNum;
    a = divFileSet->caseSampleNum;
    b = divFileSet->controlSampleNum;
    n = a + b;

    /* modified 2012.Feb setting M_Type appropiated of Markov.Type */
	M_Type=inputTypeI->CalcType;

//	M_Type=SetMType(inputTypeI->MarkovType);



/****************************************************************/
/* データ入力                                                   */
/****************************************************************/

#ifdef MPI_USE
    /* 時間計測 */
    tm[TM_B_READ] = MPI_Wtime();
#endif

    /* プロセッサごとの担当ブロック個数計算 */
    blockPartNum = blockNumAll / MyMpiSize;
    if (MyMpiRank < blockNumAll % MyMpiSize) {
        blockPartNum++;
    }
    /* 各プロセッサにおける担当ブロックのオフセット値の計算 */
    if (MyMpiRank < blockNumAll % MyMpiSize) {
        blockPartStart = blockPartNum * MyMpiRank;
    }
    else{
        blockPartStart = (blockPartNum + 1) * (blockNumAll % MyMpiSize)
                       + blockPartNum * (MyMpiRank - blockNumAll % MyMpiSize);
    }

#ifdef MPI_USE
    /* 担当ブロックが格納されるファイルと、開始・終了ブロックの計算 */
    retval = Check_BlockAreaRange(divFileSet, blockPartStart, blockPartNum,
				  &start_FileIdx, &start_blockIdx,
				  &end_fileIdx, &end_blockIdx);
    if ( retval != 0 ) {
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }

    /* 各haplotypeブロック領域内のSNP数格納用配列のメモリ確保 */
    linkSNPNum = (long*)malloc1Dim(sizeof(long), blockPartNum);
    if (NULL == linkSNPNum){
      fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }
    /* 各haplotypeブロック領域内の最初のSNPを示す座位格納用配列のメモリ確保 */
    linkSNPStart = (long*)malloc1Dim(sizeof(long), blockPartNum);
    if (NULL == linkSNPStart){
      fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }

    /* 各haplotypeブロック領域に該当するSNPデータを決定 */
    if ( Link_SNP_DivData(divFileSet, start_FileIdx, end_fileIdx,
			  start_blockIdx, end_blockIdx,
			  linkSNPStart, linkSNPNum, &dataNum) != 0 ) {
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }

    /* snpData1（case）のメモリ確保 */
    snpData1 = (SnpData *)malloc1Dim(sizeof(SnpData), dataNum);
    if (snpData1 == NULL) {
      fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }
    for ( i = 0; i < dataNum; i++ ) {
      snpData1[i].SNPdata = (char*)malloc1Dim(sizeof(char), a);
      if ( ! snpData1[i].SNPdata ) {
	fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
	MPI_Abort(MPI_COMM_WORLD, 1);
	goto finalize;
      }
    } // end of for(i)

    /* snpData2（control）のメモリ確保 */
    snpData2 = (SnpData *)malloc1Dim(sizeof(SnpData), dataNum);
    if (snpData2 == NULL) {
      fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }
    for ( i = 0; i < dataNum; i++ ) {
      snpData2[i].SNPdata = (char*)malloc1Dim(sizeof(char), b);
      if ( ! snpData2[i].SNPdata ) {
	fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
	MPI_Abort(MPI_COMM_WORLD, 1);
	goto finalize;
      }
    } // end of for(i)

    /* haplotypeブロック領域のメモリ確保 */
    if ( inputTypeI->areaFileType == 0 ) {
      areaFileLine = blockPartNum + 1;
      blockArea = (long*)malloc1Dim(sizeof(long), areaFileLine);
      if (blockArea == NULL) {
	fprintf(stderr, "rank:%d メモリを確保できませんでした.\n", MyMpiRank);
	MPI_Abort(MPI_COMM_WORLD, 1);
	goto finalize;
      }
    }


    /* 分割データの読み込み */
    if ( Read_SNP_DivData(divFileSet, start_FileIdx, end_fileIdx,
			  start_blockIdx, end_blockIdx,
			  snpData1, snpData2, blockArea) != 0 ) {
      MPI_Abort(MPI_COMM_WORLD, 1);
      goto finalize;
    }

    /* 時間計測 */
    tm[TM_A_READ] = MPI_Wtime();
    tm[TM_B_F_BCAST] = tm[TM_A_READ];
    tm[TM_A_F_BCAST] = tm[TM_A_READ];
    tm[TM_B_BCAST] = tm[TM_A_READ];
    tm[TM_A_BCAST] = tm[TM_A_READ];
#endif

    /* 偶現表Tのメモリ確保 */
    T = (int**)mallocInt2Dim(ROW, COLUMN);
    if (NULL == T){ goto finalize; }
    /* populationTypeのメモリ確保 */
    populationType = (int*)malloc1Dim(sizeof(int), n);
    if (NULL == populationType) { goto finalize; }
    /* type I error 格納配列のメモリ確保 */
    typeIError = (double*)malloc1Dim(sizeof(double), blockPartNum * repeat);
    if (NULL == typeIError) { goto finalize; }
    /* haplotypeデータのメモリ確保 */
    haplotype = (char***)malloc1Dim(sizeof(char**), blockPartNum);
    if (NULL == haplotype) { goto finalize; }
    /* パターン頻度のメモリ確保 */
    freq = (double**)malloc1Dim(sizeof(double*), blockPartNum);
    if (NULL == freq) { goto finalize; }
    /* パターン数のメモリ確保 */
    L = (long*)malloc1Dim(sizeof(long), blockPartNum);
    if (NULL == L) { goto finalize; }
    /* Haplotypeブロック内のスコア最大値の座位のメモリ確保 */
    maxScoreIndex = (long*)malloc1Dim(sizeof(long), blockPartNum);
    if (NULL == maxScoreIndex) { goto finalize; }
    /* Haplotypeブロック毎のスコア最大値のメモリ確保 */
    Sobs = (double*)malloc1Dim(sizeof(double), blockPartNum);
    if (NULL == Sobs) { goto finalize; }


/****************************************************************/
/* 検定処理                                                     */
/****************************************************************/
#ifdef MPI_USE
    /* 時間計測 */
    tm[TM_B_CALC] = MPI_Wtime();
#endif

    /* haplotypeブロック単位でループ */
    jStart = 0;
    for (i = 0, k = 0 /* blockPartStart */; i < blockPartNum; i++, k++){
        jStart = linkSNPStart[k];
        jEnd = jStart + linkSNPNum[k];
        /* 各haplotypeブロック内でループ */
        for (j = jStart; j < jEnd; j++){
            /* 観測値から偶現表を作成する */
            DataReaderPopulationType(&snpData1[j], &snpData2[j], populationType);
            DataReaderMakeTableDi(&snpData1[j], &snpData2[j], populationType, T);
            /* スコアを計算　スコア計算中0割になってしまう場合を-1を返す */
            S = TableCalcScore(T);
            /* 最大スコアを決定する */
            if (S > Sobs[i]){
                Sobs[i] = S;
                maxScoreIndex[i] = j;
            }
        }
        /* 各Haplotypeの頻度を計算 */
	if(M_Type!=0){
 	       L[i] = DataReaderHaplotypeFrequency(snpData1, snpData2, jStart, jEnd, &haplotype[i], &freq[i]);
	}

        /* ハプロタイプ関連解析のType I errorの確率計算を実行する */
	/* modified 2012.Feb add parameter M_Type for selectable Markov calculation */
        if (Sobs[i] > 0.0) {
//            TypeIExecute(freq[i], haplotype[i], L[i], Sobs[i], a, b, linkSNPNum[k], &typeIError[i*repeat], repeat, inputTypeI->gen,M_Type);

		if(M_Type==0){
			RATExecute(snpData1,snpData2,Sobs[i],a,b,jStart,jEnd,&typeIError[i*repeat],repeat,inputTypeI->gen,flag);
		}else{	
			TypeIExecuteHybridMFast(freq[i],haplotype[i],L[i],Sobs[i],a,b,linkSNPNum[k],&typeIError[i*repeat],repeat,inputTypeI->gen,M_Type);
		}
        }
    }

#ifdef MPI_USE
    /* 時間計測 */
    tm[TM_A_CALC] = MPI_Wtime();

    if (MyMpiRank == 0) {
        /* 各ランクのブロック数を格納する配列のメモリを確保 */
        numOutputs = (int *)malloc1Dim(sizeof(int), MyMpiSize);
        if (numOutputs == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
        /* 各ランクデータのオフセット値を格納する配列のメモリを確保 */
        outDataOffsets = (int *)malloc1Dim(sizeof(int), MyMpiSize);
        if (outDataOffsets == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
        /* 各ランクのTypeIerror計算結果数を格納する配列のメモリを確保 */
        numTIEOutputs = (int *)malloc1Dim(sizeof(int), MyMpiSize);
        if (numTIEOutputs == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
        /* 各ランクのTypeIerro計算結果のオフセット値を格納するメモリを確保 */
        outTIEDataOffsets = (int *)malloc1Dim(sizeof(int), MyMpiSize);
        if (outTIEDataOffsets == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
    }

    /* 時間計測 */
    tm[TM_B_F_GATHER] = MPI_Wtime();

    numOutput = blockPartNum;

    /* 各ランクが計算した出力値の個数 */
    MPI_Gather(&numOutput, 1, MPI_INT, numOutputs, 1,MPI_INT, 0, MPI_COMM_WORLD);

    /* 時間計測 */
    tm[TM_A_F_GATHER] = MPI_Wtime();

    if (MyMpiRank == 0) {
        /* TypeIErrorの個数設定 */
        for (i = 0; i < MyMpiSize; i++) {
            numTIEOutputs[i] = numOutputs[i] * repeat;
        }

        /* ランクごとのデータのオフセット値の計算 */
        outDataOffsets[0] = 0;
        outTIEDataOffsets[0] = 0;
        for (i = 1; i < MyMpiSize; i++) {
            outDataOffsets[i] = outDataOffsets[i-1] + numOutputs[i-1];
            outTIEDataOffsets[i] = outTIEDataOffsets[i-1] + numOutputs[i-1] * repeat;
        }

        /* 出力用データを格納する構造体配列のメモリ確保 */
        SobsAll = (double *)malloc1Dim(sizeof(double), blockNumAll);
        maxScoreIndexAll = (long *)malloc1Dim(sizeof(long), blockNumAll);
        typeIErrorAll = (double *)malloc1Dim(sizeof(double), blockNumAll * repeat);
	linkSNPNumAll = (long *)malloc1Dim(sizeof(long), blockNumAll);
	linkSNPStartAll = (long *)malloc1Dim(sizeof(long), blockNumAll);
	blockAreaAll = (long *)malloc1Dim(sizeof(long), blockNumAll + 1);
    }

    /* 時間計測 */
    tm[TM_B_GATHER] = MPI_Wtime();

    /* 各プロセッサからのデータの収集 */
    MPI_Gatherv(Sobs, numOutput, MPI_DOUBLE, SobsAll,
               numOutputs, outDataOffsets, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Gatherv(maxScoreIndex, numOutput, MPI_LONG, maxScoreIndexAll,
               numOutputs, outDataOffsets, MPI_LONG, 0, MPI_COMM_WORLD);
    MPI_Gatherv(typeIError, numOutput * repeat, MPI_DOUBLE, typeIErrorAll,
               numTIEOutputs, outTIEDataOffsets, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    MPI_Gatherv(linkSNPNum, numOutput, MPI_LONG, linkSNPNumAll,
               numOutputs, outDataOffsets, MPI_LONG, 0, MPI_COMM_WORLD);
    if ( MyMpiRank == 0 ) {
      linkSNPStartAll[0] = 0;
      for ( i = 1; i < blockNumAll; i++ )
	linkSNPStartAll[i] = linkSNPStartAll[i-1] + linkSNPNumAll[i-1];
    }
    if ( MyMpiRank == 0 )
      numOutputs[MyMpiSize - 1] += 1;
    if ( MyMpiRank == MyMpiSize - 1 )
      MPI_Gatherv(blockArea, numOutput + 1, MPI_LONG, blockAreaAll,
		  numOutputs, outDataOffsets, MPI_LONG, 0, MPI_COMM_WORLD);
    else
      MPI_Gatherv(blockArea, numOutput, MPI_LONG, blockAreaAll,
		  numOutputs, outDataOffsets, MPI_LONG, 0, MPI_COMM_WORLD);


    /* 時間計測 */
    //tm[TM_A_GATHER] = MPI_Wtime();


    /* 各ランクが処理したデータの個数 */
    numData = dataNum;
    if (MyMpiRank == 0) {
        numDataAll = (int *)malloc1Dim(sizeof(int), MyMpiSize);
	if ( numDataAll == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
    }
    MPI_Gather(&numData, 1, MPI_INT, numDataAll, 1,MPI_INT, 0, MPI_COMM_WORLD);

    /* 送信用データ */
    snpData_rsN = (char *)malloc1Dim(sizeof(char), numData * RS_NUM_LEN);
    snpData_pos = (long *)malloc1Dim(sizeof(long), numData);
    for (i = 0; i < numData; i++ ) {
        memcpy(&snpData_rsN[i*RS_NUM_LEN], snpData1[i].rsNumber, sizeof(char)*RS_NUM_LEN);
	snpData_pos[i] = snpData1[i].pos;
    } // end of for(i)

    if (MyMpiRank == 0) {
        for ( i = 0; i < MyMpiSize; i++ ) numDataTotal += numDataAll[i];

        /* 受信用データ */
	snpDataAll_rsN = (char *)malloc1Dim(sizeof(char), numDataTotal * RS_NUM_LEN);
	snpDataAll_pos = (long *)malloc1Dim(sizeof(long), numDataTotal);

        /* 受信用オフセットデータ */
	outSnpRsNNums = (int *)malloc1Dim(sizeof(int), MyMpiSize);
	outSnpPosNums = (int *)malloc1Dim(sizeof(int), MyMpiSize);
	outSnpRsNOffsets = (int *)malloc1Dim(sizeof(int), MyMpiSize);
	outSnpPosOffsets = (int *)malloc1Dim(sizeof(int), MyMpiSize);
        if (outSnpRsNNums == NULL || outSnpPosNums == NULL ||
	    outSnpRsNOffsets == NULL || outSnpPosOffsets == NULL) {
            MPI_Abort(MPI_COMM_WORLD, 1);
            goto finalize;
        }
	for (i = 0; i < MyMpiSize; i++) {
	    outSnpRsNNums[i] = numDataAll[i] * RS_NUM_LEN;
	    outSnpPosNums[i] = numDataAll[i];
	}
	outSnpRsNOffsets[0] = 0;
	outSnpPosOffsets[0] = 0;
        for (i = 1; i < MyMpiSize; i++) {
	    outSnpRsNOffsets[i] = outSnpRsNOffsets[i-1] + numDataAll[i-1] * RS_NUM_LEN;
	    outSnpPosOffsets[i] = outSnpPosOffsets[i-1] + numDataAll[i-1];
	} // end of for(i)
    }

    /* 各プロセッサからのデータの収集 */
    MPI_Gatherv(snpData_rsN, dataNum*RS_NUM_LEN, MPI_CHAR, snpDataAll_rsN,
		outSnpRsNNums, outSnpRsNOffsets, MPI_CHAR, 0, MPI_COMM_WORLD);
    MPI_Gatherv(snpData_pos, dataNum, MPI_LONG, snpDataAll_pos,
		outSnpPosNums, outSnpPosOffsets, MPI_LONG, 0, MPI_COMM_WORLD);

    /* maxScoreIndexAll のインデックス値の修正 */
    if (MyMpiRank == 0) {
        numOutputs[MyMpiSize - 1] -= 1;
        index = numOutputs[0];
	countData = numDataAll[0];
	for ( i = 1; i < MyMpiSize; i++ ) {
	  for ( j = 0; j < numOutputs[i]; j++ ) {
	    maxScoreIndexAll[index++] += countData;
	  }
	  countData += numDataAll[i];
	}
    }

    /* 時間計測 */
    tm[TM_A_GATHER] = MPI_Wtime();


    /* 時間計測 */
    tm[TM_B_WRITE] = MPI_Wtime();
#endif

/****************************************************************/
/* 検定結果出力                                                 */
/****************************************************************/
    if (MyMpiRank == 0) {
        retval = OutputFileOpen(&fpOut, inputTypeI->outputFile1);
        if (retval != 0){
		printf("outputFile open error\n");
            goto finalize;
        }

        fprintf(fpOut, "CaseData    = %s\n", inputTypeI->inputFile1);
        fprintf(fpOut, "ControlData = %s\n", inputTypeI->inputFile2);
        fprintf(fpOut, "Repeat      = %ld\n", inputTypeI->repeat);
        fprintf(fpOut, "Genoration  = %ld\n", inputTypeI->gen);
	/* modified 2012.Feb add print Markov calculation type */

	if(inputTypeI->CalcType==0){
		fprintf(fpOut,"RAT Calculation\n\n");
	}else{
		fprintf(fpOut,"Markov Calculation : M_Type=%d\n\n",inputTypeI->CalcType);
	}
//	fprintf(fpOut,"Markov calculation Type = %s \n",inputTypeI->MarkovType);

        fprintf(fpOut, "BlockArea\tSNPNum\trsNumber\tPosition\tScore\tTypeIerror...\n");
 
        if (inputTypeI->areaFileType == 0){
            for (i = 0; i < blockNumAll; i++){
                if (SobsAll[i] != 0){
                    fprintf(fpOut, "%ld-%ld\t%ld\t%s\t%ld\t%.3e\t", 
                            blockAreaAll[i],
                            blockAreaAll[i+1],
                            linkSNPNumAll[i],
                            &snpDataAll_rsN[ maxScoreIndexAll[i] * RS_NUM_LEN ],
                            snpDataAll_pos[ maxScoreIndexAll[i] ],
                            SobsAll[i]);
                    /* type I errorを出力 */
                    for (index = 0; index < repeat; index++){
                        fprintf(fpOut, "%.3e\t", typeIErrorAll[i*repeat+index]);
                    }
                    fprintf(fpOut, "\n");
                }
                else {
                    fprintf(fpOut, "%ld-%ld\t%ld\tNoData\n", 
                            blockAreaAll[i],
                            blockAreaAll[i+1],
                            linkSNPNumAll[i]);
                }
            }
        }
        else {
            for (i = 0; i < blockNumAll; i++){
                if (SobsAll[i] != 0){
                    fprintf(fpOut, "%ld-%ld\t%ld\t%s\t%ld\t%.3e\t", 
                            snpDataAll_pos[ linkSNPStartAll[i] ],
                            snpDataAll_pos[ linkSNPStartAll[i] + linkSNPNumAll[i] - 1 ],
                            linkSNPNumAll[i],
                            &snpDataAll_rsN[ maxScoreIndexAll[i] * RS_NUM_LEN ],
                            snpDataAll_pos[ maxScoreIndexAll[i] ],
                            SobsAll[i]);
                    /* type I errorを出力 */
                    for (index = 0; index < repeat; index++){
                        fprintf(fpOut, "%.3e\t", typeIErrorAll[i*repeat+index]);
                    }
                    fprintf(fpOut, "\n");
                }
                else {
                    fprintf(fpOut, "%ld-%ld\t%ld\tNoData\n", 
                            snpDataAll_pos[ linkSNPStartAll[i] ],
                            snpDataAll_pos[ linkSNPStartAll[i] + linkSNPNumAll[i] - 1 ],
                            linkSNPNumAll[i]);
                }
            }
        }
    }

#ifdef MPI_USE
    /* 時間計測 */
    tm[TM_A_WRITE] = MPI_Wtime();
#endif
/****************************************************************/
/* 終了処理                                                     */
/****************************************************************/

finalize:;
    /* ファイルクローズ */
    FileClose(fpOut);

    /* 確保したメモリを開放する */
    // haplotype変数はchar型の３次元配列
    // 各列の要素数は配列Lからわかる。文字数はlinkedSNPNumである。
    // 専用の開放の関数を作成する必要がある。とりあえず普通にfree
    for (i = blockPartNum-1; i >= 0; i--){
        for (index = L[i]-1; index >= 0; index--){
            free1Dim(haplotype[i][index]);
        }
        free1Dim(haplotype[i]);
    }
    free1Dim(haplotype);

    free1Dim(numOutputs);
    free1Dim(outDataOffsets);
    free1Dim(outTIEOutputs);
    free1Dim(outTIEDataOffsets);
    free1Dim(numTIEOutputs);

    free1Dim(SobsAll);
    free1Dim(typeIErrorAll);
    free1Dim(maxScoreIndexAll);
    free1Dim(blockAreaAll);

    free1Dim(blockArea);
    free1Dim(linkSNPNum);
    free1Dim(linkSNPStart);
    free1Dim(typeIError);
    freeDouble2Dim(freq,blockPartNum);
    freeInt2Dim(T, ROW);
    free1Dim(populationType);
    free1Dim(L);
    free1Dim(maxScoreIndex);
    free1Dim(Sobs);
    DataReaderSnpDataMemoryFree(snpData1, dataNum);
    DataReaderSnpDataMemoryFree(snpData2, dataNum);

    return;
}



/* デバッグ用 */
void CheckCopySNPData(SnpData *snpData, long dataNum, int MyMpiRank, int IDNum)
{
    long i = 0, j = 0;
    static long len = 0;
    static char *buf = NULL;

    if (len < 22 + snpData[i].dataNum * 2 + 2) {
        len = 22 + snpData[i].dataNum * 2 + 2;
        if (buf != NULL)  free(buf);
        buf = (char*)malloc(sizeof(char) * len);
        for (i = 0; i < len; i++)  buf[i] = '\0';
    }

//    printf("SNP Data  rank  dNum\n");
    for (i = 0; i < dataNum; i++) {
        printf("%-4d  %-4d  %-6ld    ", IDNum, MyMpiRank, i); // 22文字
        for (j = 0; j < snpData[i].dataNum; j+=2) {
            printf("%c%c  ", snpData[i].SNPdata[j],
                             snpData[i].SNPdata[j+1]);
        }
        printf("\n");
    }

	return;
}
