/*!
  @file      snp_typeI_DivFile.c
  @brief     typeIDivFile.exe Main program file
  @attention based on snp_typeI.c
  @date      March 2011
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>
#include <libgen.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>

#define _MAIN_DEF

#include "snp_typeI.h"


/* プロトタイプ宣言 */
int MainProgramDiv(InputTypeI *inputTypeI, char *OutputBase);

int OutputDivFiles(char *OutputBase, long dataNum, long blockNum,
		   long *blockPartStart, long *blockPartNum,
		   long *linkSNPStart, long *linkSNPNum,
		   SnpData *caseSnpData, long caseSampleNum,
		   SnpData *controlSnpData, long controlSampleNum,
		   long *blockArea, long blockAreaNum);

int OutputDivXml(InputTypeI *inputTypeI, char *OutputBase,
		 long dataNum, long blockNum,
		 long *blockPartStart, long *blockPartNum,
		 long caseSampleNum, long controlSampleNum);


/*====================================================================*/
/* メイン関数 */
/*====================================================================*/
int main(int argc, char* argv[])
{
  InputTypeI inputTypeI={"", "", "", 0, "", 0, 0, 0, 0};
  char OutputBase[MAX_LEN] = {'\0',};
  int i = 0;
  if ( argc != 8 ) {
    printf("[usage]%s [DataType] [InputFile1] [InputFile2]"
	   " [AreaFileType] [BlockAreaFile] [DivNum] [OutputBase] \n",
	   argv[0]);
    return 255;
  }
  inputTypeI.dataType = atoi(argv[1]);
  strcpy(inputTypeI.inputFile1, argv[2]);
  strcpy(inputTypeI.inputFile2, argv[3]);
  inputTypeI.areaFileType = atoi(argv[4]);
  strcpy(inputTypeI.blockAreaFile, argv[5]);
  MyMpiSize = atoi(argv[6]); // 分割数(DivNum)はグローバル変数MyMpiSizeに格納
  strcpy(OutputBase, argv[7]);
  MyMpiRank = 0;

  printf("inputFile1: %s\n", inputTypeI.inputFile1);
  printf("inputFile2: %s\n", inputTypeI.inputFile2);
  printf("outpuBase: %s\n", OutputBase);
  printf("blockAreaFile: %s\n", inputTypeI.blockAreaFile);
  printf("dataType: %d\n", inputTypeI.dataType);
  printf("divNum: %d\n", MyMpiSize);

  /* ファイル分割を行う */
  MainProgramDiv(&inputTypeI, OutputBase);

  return 0;
}


/*====================================================================*/
/*! ファイル分割主処理
  @param   inputTypeI   入力データ格納用構造体へのポインタ
  @param   OutputBase   出力ファイルのベース名
 */
/*====================================================================*/
int MainProgramDiv(InputTypeI *inputTypeI, char *OutputBase)
{
  int retval = 0;
  int flag = 0;
  long i = 0;
  long j = 0;
  long k = 0;
  long fileLine1 = 0; /* 入力ファイルのライン数 */
  long fileLine2 = 0; /* 入力ファイルのライン数 */
  long areaFileLine = 0; /* haplotypeブロック領域指定ファイルのライン数 */
  long a = 0;         /* number of haplotype copies (=sequences) in case */
  long b = 0;         /* number of haplotype copies (=sequences) in control */
  long dataNum;       /* case、controlデータ整合後の総SNP数 */
  long blockNum = 0;  /* haplotypeブロック数 */
  long startPos = 0;
  long endPos = 0;
  long *blockPartNum = NULL;
  long *blockPartStart = NULL;

  char *caseData = NULL;       /* サンプルデータ（case）格納用 */
  char *controlData = NULL;    /* サンプルデータ（control）格納用 */
  long *blockArea = NULL;      /* 各haplotypeブロック領域格納 */
  long *linkSNPNum = NULL;     /* 各haplotypeブロックのSNP数 */
  long *linkSNPStart = NULL;   /* 各haplotypeブロックの最初のSNPを示す座位 */

  FILE *fpCase = NULL;         /* 入力（case）ファイルポインタ */
  FILE *fpCntl = NULL;         /* 入力（control）ファイルポインタ */
  FILE *fpArea = NULL;         /* haplotypeブロック領域指定ファイルポインタ */

  SnpData *snpTmpData1 = NULL;
  SnpData *snpTmpData2 = NULL;
  SnpData *snpData1 = NULL;
  SnpData *snpData2 = NULL;


  /****************************************************************/
  /* データ入力                                                   */
  /****************************************************************/

  /* ファイルオープン */
  retval = InputFileOpen(&fpCase, inputTypeI->inputFile1);
  if (retval != 0){
    goto finalize;
  }
  retval = InputFileOpen(&fpCntl, inputTypeI->inputFile2);
  if (retval != 0){
    goto finalize;
  }
  retval = InputFileOpen(&fpArea, inputTypeI->blockAreaFile);
  if (retval != 0){
    goto finalize;
  }

  /* 入力ファイルのライン数を取得 */
  fileLine1 = DataReaderCountFileLine(fpCase);
  fileLine2 = DataReaderCountFileLine(fpCntl);

  /* ファイルポインタを先頭に戻す */
  fseek(fpCase, 0L, SEEK_SET);
  fseek(fpCntl, 0L, SEEK_SET);

  /* データ一時格納用構造体のメモリ確保 */
  snpTmpData1 = (SnpData*)malloc1Dim(sizeof(SnpData), fileLine1);
  if (NULL == snpTmpData1){ retval = -1; goto finalize; }
  snpTmpData2 = (SnpData*)malloc1Dim(sizeof(SnpData), fileLine2);
  if (NULL == snpTmpData2){ retval = -1; goto finalize; }

  /* データをファイルから読み込み構造体に収める */
  DataReaderSetAllData(fpCase, snpTmpData1, fileLine1, inputTypeI->dataType);
  DataReaderSetAllData(fpCntl, snpTmpData2, fileLine2, inputTypeI->dataType);

  /* 入力データの整合性をチェックして並列化用の入力データを作成する */
  /* サンプルデータは別配列（caseData, controlData）で保持する */
  dataNum = DataReaderMakeParallelData(snpTmpData1, snpTmpData2,
				       fileLine1, fileLine2,
				       &snpData1, &snpData2,
				       &caseData, &controlData);

  /* 使用しない配列のメモリ開放 */
  /* 構造体SnpDataメンバのメモリを開放する */
  DataReaderSnpDataMemoryFree(snpTmpData1, fileLine1);
  DataReaderSnpDataMemoryFree(snpTmpData2, fileLine2);
  snpTmpData1 = NULL;
  snpTmpData2 = NULL;
 
  /* haplotypeブロック領域指定ファイルのライン数を取得 */
  areaFileLine = DataReaderCountFileLine(fpArea);
  /* haplotypeブロック領域格納用配列のメモリ確保 */
  blockArea = (long*)malloc1Dim(sizeof(long), areaFileLine);
  if (NULL == blockArea){ retval = -1; goto finalize; }
  /* ファイルポインタを先頭に戻す */
  fseek(fpArea, 0L, SEEK_SET);
  /* haplotypeブロック領域を配列に収める */
  DataReaderSetHaplotypeBlockArea(fpArea, blockArea);
  /* haplotypeブロック数 */
  if (inputTypeI->areaFileType == 0){
    blockNum = areaFileLine - 1;
  }
  else {
    /* haplotypeブロックの個数を計算する */
    if (blockArea[0] > dataNum){
      blockNum = 1;
      blockArea[1] = dataNum; /* linkSNPNumの値がデータ数だけになる */
    }
    else{
      blockNum = (dataNum - (blockArea[0] - blockArea[1]) ) / blockArea[1];
    }
  }

  if (MyMpiSize > blockNum) {
    fprintf(stderr,
	    "分割数(%d)がデータブロック数(%ld)よりも多く設定されています.\n",
	    MyMpiSize, blockNum);
    retval = -1; goto finalize;
  }
  if (MyMpiSize > dataNum) {
    fprintf(stderr,
	    "分割数(%d)がデータ数(%ld)よりも多く設定されています.\n",
	    MyMpiSize, dataNum);
    retval = -1; goto finalize;
  }

  /* 入力データのサンプル数取得 */
  a = snpData1[0].dataNum;
  b = snpData2[0].dataNum;
  //n = a + b;

  /* 構造体SnpDataにサンプルデータをコピーする */
  DataReaderDataCopyToSnpData(snpData1, caseData, dataNum, a);
  DataReaderDataCopyToSnpData(snpData2, controlData, dataNum, b);


  /****************************************************************/
  /* メモリ確保                                                   */
  /****************************************************************/

  /* プロセッサごとのデータ個数計算 */
  blockPartNum = (long*)malloc1Dim(sizeof(long), MyMpiSize);
  if (NULL == blockPartNum){ retval = -1; goto finalize; }
  blockPartStart = (long*)malloc1Dim(sizeof(long), MyMpiSize);
  if (NULL == blockPartStart){ retval = -1; goto finalize; }
    
  for ( i = 0; i < MyMpiSize; i++ ) {
    blockPartNum[i] = blockNum / MyMpiSize;
    if (i < blockNum % MyMpiSize) {
      blockPartNum[i]++;
    }
    /* 各プロセッサにおけるデータのオフセット値の計算 */
    if (i < blockNum % MyMpiSize) {
      blockPartStart[i] = blockPartNum[i] * i;
    }
    else{
      blockPartStart[i] = (blockPartNum[i] + 1) * (blockNum % MyMpiSize)
	+ blockPartNum[i] * (i - blockNum % MyMpiSize);
    }
  } // end of for(i)

  /* 各haplotypeブロック領域内のSNP数格納用配列のメモリ確保 */
  linkSNPNum = (long*)malloc1Dim(sizeof(long), blockNum);
  if (NULL == linkSNPNum){ retval = -1; goto finalize; }
  /* 各haplotypeブロック領域内の最初のSNPを示す座位格納用配列のメモリ確保 */
  linkSNPStart = (long*)malloc1Dim(sizeof(long), blockNum);
  if (NULL == linkSNPStart){ retval = -1; goto finalize; }

  /* 各haplotypeブロック領域に該当するSNPデータを決定 */
  if (inputTypeI->areaFileType == 0){
    for (i = 0; i < blockNum; i++){ /* 将来、領域の重複を許す場合も考える */
      startPos = blockArea[i];
      endPos = blockArea[i+1];
      flag = 0;
      /* 入力データはポジションでソートされていると仮定 */
      for (j = 0; j < dataNum; j++){
	if (startPos <= snpData1[j].pos){
	  if (snpData1[j].pos < endPos){
	    /* 領域内のSNP数カウント */
	    linkSNPNum[i]++;
	    /* 領域内の最初のSNPを示す座位を保持 */
	    if (0 == flag ){
	      linkSNPStart[i] = j;
	      flag = 1;
	    }
	  }
	  /* これ以降、領域に該当するデータは出現しないので
	     次のブロックを調べる */
	  else {
	    break;
	  }
	}
      } // end of for(j)
    } // end of for(i)
  }
  else {
    for (i = 0; i < blockNum; i++){
      linkSNPNum[i] = blockArea[0];
      linkSNPStart[i] = i * blockArea[1];
    } // end of for(i)
  }

  /****************************************************************/
  /* ファイル出力処理                                             */
  /****************************************************************/

  /* 分割データファイル群の出力 */
  if ( inputTypeI->areaFileType == 0 ) {
    if ( OutputDivFiles(OutputBase, dataNum, blockNum,
			blockPartStart, blockPartNum,
			linkSNPStart, linkSNPNum,
			snpData1, a, snpData2, b,
			blockArea, areaFileLine) != 0 ) {
      retval = -1; goto finalize;
    }
  } else {
    if ( OutputDivFiles(OutputBase, dataNum, blockNum,
			blockPartStart, blockPartNum,
			linkSNPStart, linkSNPNum,
			snpData1, a, snpData2, b,
			NULL, 0) != 0 ) {
      retval = -1; goto finalize;
    }
  }

  /* 分割データ情報XMLファイルの出力 */
  if ( OutputDivXml(inputTypeI, OutputBase, dataNum, blockNum,
		    blockPartStart, blockPartNum, a, b) != 0 ) {
    retval = -1; goto finalize;
  }

  /****************************************************************/
  /* 終了処理                                                     */
  /****************************************************************/

 finalize:;
  /* ファイルクローズ */
  FileClose(fpCase);
  FileClose(fpCntl);
  FileClose(fpArea);

  /* メモリ開放 */
  free1Dim(blockArea);
  free1Dim(linkSNPNum);
  free1Dim(linkSNPStart);
  DataReaderSnpDataMemoryFree(snpTmpData1, fileLine1);
  DataReaderSnpDataMemoryFree(snpTmpData2, fileLine2);
  free1Dim(caseData);
  free1Dim(controlData);
  free1Dim(snpData1);
  free1Dim(snpData2);

  free1Dim(blockPartNum);
  free1Dim(blockPartStart);

  if ( retval != 0 ) return -1; 
  return 0;
}

/*====================================================================*/
/*! 出力ディレクトリチェック
  @param   OutputBase   出力ファイルのベース名
  @attention dirname(OutputBase)のディレクトリが存在しない場合、作成します。
 */
/*====================================================================*/
static int Util_CheckDir(char *OutputBase) {
  char *outDir, wkStr[MAX_LEN];

  if ( ! OutputBase || strlen(OutputBase) < 1 )
    return -1;

  strcpy(wkStr, OutputBase);
  outDir = dirname(wkStr);
  if ( outDir ) {
    DIR *dirp;
    dirp = opendir(outDir);
    if ( ! dirp ) {
      char comm[MAX_LEN];
      switch ( errno ) {
      case ENOENT:
	sprintf(comm, "mkdir -p %s", outDir);
	if ( system(comm) != 0 ) return -1;
	dirp = opendir(outDir);
	if ( ! dirp ) {
	  perror(outDir);
	  return -1;
	}
	else closedir(dirp);
	break;
      default:
	perror(outDir);
	return -1;
      }
      return 0;
    }
    closedir(dirp);
    return 0;
  }
  // dir is NULL, somethin wrong...
  return -1;
}

/*====================================================================*/
/*! SNPデータレコードバイナリ出力処理
  @param   fp        出力先のファイルポインタ
  @param   data      出力するSNPデータ構造体へのポインタ
  @param   sampleNum SNPデータのサンプル数
 */
/*====================================================================*/
int DumpSNPdata(FILE *fp, SnpData* data, long sampleNum)
{
  if ( ! fp || ! data ) return -1;
  if ( fwrite(data->rsNumber, sizeof(char), RS_NUM_LEN, fp) != RS_NUM_LEN )
    return -1;
  if ( fwrite(data->SNPalleles, sizeof(char), SNP_NUM, fp) != SNP_NUM )
    return -1;
  if ( fwrite(data->chrom, sizeof(char), CHROM_LEN, fp) != CHROM_LEN )
    return -1;
  if ( fwrite(&data->pos, sizeof(long), 1, fp) != 1 )
    return -1;
  if ( fwrite(&data->allelesNum, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( fwrite(&data->dataNum, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( fwrite(data->SNPallelesCount, sizeof(int), SNP_COUNT, fp) != SNP_COUNT )
    return -1;
  if ( fwrite(&data->missingDataCount, sizeof(int), 1, fp) != 1 )
    return -1;
  if ( sampleNum > 0 ) {
    if ( fwrite(data->SNPdata, sizeof(char), sampleNum, fp) != sampleNum )
      return -1;
  }
  return 0;
}

/*====================================================================*/
/*! 分割データファイル出力処理
  @param   OutputBase     出力ファイルのベース名
  @param   dataNum        総SNPデータ数
  @param   blockNum       総ブロック数
  @param   blockPartStart 担当するブロック領域のデータ全体における開始位置リスト
  @param   blockPartNum   担当するブロック領域の総数のリスト
  @param   linkSNPStart   ブロック領域毎のデータ開始位置リスト
  @param   linkSNPNum     ブロック領域毎のデータ数のリスト
  @param   caseSnpData    CASEデータ格納領域へのポインタ
  @param   caseSampleNum  CASEデータのサンプル数
  @param   controlSnpData CONTROLデータ格納領域へのポインタ
  @param   controlSampleNum  CONTROLデータのサンプル数
  @param   blockArea      ブロック領域データ格納領域へのポインタ
  @param   blockAreaNum   ブロック領域データ格納数(blockNum+1)
  @attention blockAreaNumが0の場合、areaFileType!=0としてブロック領域データは出力されません。
  @attention 作成されるファイルは「OutputBase_番号.dat」のファイル名になります。
 */
/*====================================================================*/
int OutputDivFiles(char *OutputBase, long dataNum, long blockNum,
		   long *blockPartStart, long *blockPartNum,
		   long *linkSNPStart, long *linkSNPNum,
		   SnpData *caseSnpData, long caseSampleNum,
		   SnpData *controlSnpData, long controlSampleNum,
		   long *blockArea, long blockAreaNum)
{
  int divNum = MyMpiSize;
  int divNumOfDigit;
  char outFileName[MAX_LEN], divNumTmpl[64];
  int blockCount, linkCount, snpDataCount;
  int i, j, k;

  if ( divNum <= 0 ) return -1;
  divNumOfDigit = (int)log10((double)divNum) + 1;
  sprintf(divNumTmpl, "%%s_%%0%dd.dat", divNumOfDigit);

  if ( Util_CheckDir(OutputBase) != 0 )
    return -1;

  for ( k = 0; k < divNum; k++ ) {
    FILE *fp = NULL;
    long headerWk[2];
    long totalSNPNum;

    sprintf(outFileName, divNumTmpl, OutputBase, k);
    printf("%d: %s\n", k, outFileName);
    fp = fopen(outFileName, "wb");
    if ( ! fp ) {
      perror(outFileName);
      return -1;
    }

    // blockNum
    if ( fwrite(&blockPartNum[k], sizeof(long), 1, fp) != 1 ) {
      perror(outFileName);
      fclose(fp);
      return -1;
    }

    // SNPStart, SNPNum
    totalSNPNum = 0;
    for ( j = 0; j < blockPartNum[k]; j++ ) {
      headerWk[0] = linkSNPStart[blockPartStart[k] + j]
	- linkSNPStart[blockPartStart[k]];
      headerWk[1] = linkSNPNum[blockPartStart[k] + j];
      if ( fwrite(headerWk, sizeof(long), 2, fp) != 2 ) {
	perror(outFileName);
	fclose(fp);
	return -1;
      }
      totalSNPNum += linkSNPNum[blockPartStart[k] + j];
    } // end of for(j)

    // totalSNPNum
    if ( fwrite(&totalSNPNum, sizeof(long), 1, fp) != 1 ) {
      perror(outFileName);
      fclose(fp);
      return -1;
    }

    //---------- case data ----------
    // caseSampleNum
    if ( fwrite(&caseSampleNum, sizeof(long), 1, fp) != 1 ) {
      perror(outFileName);
      fclose(fp);
      return -1;
    }

    // case_SNPdata
    for ( j = 0; j < blockPartNum[k]; j++ ) {
      for ( i = 0; i < linkSNPNum[blockPartStart[k] + j]; i++ ) {
	if ( DumpSNPdata(fp, &caseSnpData[linkSNPStart[blockPartStart[k] + j] + i],
			 caseSampleNum) != 0 ) {
	  fclose(fp);
	  return -1;
	}
      } // end of for(i)
    } // end of for(j)

    //---------- control data ----------
    // controlSampleNum
    if ( fwrite(&controlSampleNum, sizeof(long), 1, fp) != 1 ) {
      perror(outFileName);
      fclose(fp);
      return -1;
    }

    // control_SNPdata
    for ( j = 0; j < blockPartNum[k]; j++ ) {
      for ( i = 0; i < linkSNPNum[blockPartStart[k] + j]; i++ ) {
	if ( DumpSNPdata(fp, &controlSnpData[linkSNPStart[blockPartStart[k] + j] + i],
			 controlSampleNum) != 0 ) {
	  fclose(fp);
	  return -1;
	}
      } // end of for(i)
    } // end of for(j)

    //---------- blockArea data ----------
    if ( blockAreaNum > 0 ) {
      long baNum = blockPartNum[k] + 1;
      if ( fwrite(&baNum, sizeof(long), 1, fp) != 1 ) {
	perror(outFileName);
	fclose(fp);
	return -1;
      }
      for ( j = 0; j < baNum; j++ ) {
	if ( fwrite(&blockArea[blockPartStart[k] +j], sizeof(long), 1, fp) != 1 ) {
	  perror(outFileName);
	  fclose(fp);
	  return -1;
	}
      } // end of for(j)
    } else {
      if ( fwrite(&blockAreaNum, sizeof(long), 1, fp) != 1 ) {
	perror(outFileName);
	fclose(fp);
	return -1;
      }
    }

    fclose(fp);
  } // end of for(k)

  return 0;
}

/*====================================================================*/
/*! データ分割情報XMLファイル出力処理
  @param   inputTypeI     入力データ格納用構造体へのポインタ
  @param   OutputBase     出力ファイルのベース名
  @param   dataNum        総SNPデータ数
  @param   blockNum       総ブロック数
  @param   blockPartStart 担当するブロック領域のデータ全体における開始位置リスト
  @param   blockPartNum   担当するブロック領域の総数のリスト
  @param   caseSampleNum  CASEデータのサンプル数
  @param   controlSampleNum  CONTROLデータのサンプル数
  @attention 作成されるファイルは「OutputBase.xml」のファイル名になります。
 */
/*====================================================================*/
int OutputDivXml(InputTypeI *inputTypeI, char *OutputBase,
		 long dataNum, long blockNum,
		 long *blockPartStart, long *blockPartNum,
		 long caseSampleNum, long controlSampleNum)
{
  FILE *fp = NULL;
  char *ts1 = "  ";
  char *ts2 = "    ";
  char *refBaseName, *outBaseName, wkStr[MAX_LEN];
  int divNum = MyMpiSize;
  int divNumOfDigit;
  char outFileName[MAX_LEN], divNumTmpl[64];
  int k;

  if ( ! inputTypeI || ! OutputBase ) return -1;
  if ( divNum <= 0 ) return -1;
  divNumOfDigit = (int)log10((double)divNum) + 1;
  sprintf(divNumTmpl, "%%s_%%0%dd.dat", divNumOfDigit);

#if 0 // no need to check dir
  if ( Util_CheckDir(OutputBase) != 0 )
    return -1;
#endif

  sprintf(outFileName, "%s.xml", OutputBase);
  fp = fopen(outFileName, "w");
  if ( ! fp ) return -1;

  // xml header
  fprintf(fp, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");

  // top-level node
  fprintf(fp, "<ParaHaplo_DivData>\n");

  // information nodes
  fprintf(fp, "%s<program>typeIParallelD</program>\n", ts1);

  strcpy(wkStr, inputTypeI->inputFile1);
  refBaseName = basename(wkStr);
  fprintf(fp, "%s<caseData type=\"%d\" sampleNum=\"%ld\">%s</caseData>\n",
	  ts1, inputTypeI->dataType, caseSampleNum, refBaseName);

  strcpy(wkStr, inputTypeI->inputFile2);
  refBaseName = basename(wkStr);
  fprintf(fp, "%s<controlData type=\"%d\" sampleNum=\"%ld\">%s</controlData>\n", 
	  ts1, inputTypeI->dataType, controlSampleNum, refBaseName);

  strcpy(wkStr, inputTypeI->blockAreaFile);
  refBaseName = basename(wkStr);
  fprintf(fp, "%s<blockAreaData type=\"%d\">%s</blockAreaData>\n",
	  ts1, inputTypeI->areaFileType, refBaseName);

  // dataNum node
  fprintf(fp, "%s<dataNum>%ld</dataNum>\n", ts1, dataNum);

  // blockNum node
  fprintf(fp, "%s<blockNum>%ld</blockNum>\n", ts1, blockNum);

  // 分割データファイル
  fprintf(fp, "%s<SNP_Files divNum=\"%d\" baseDir=\".\">\n", ts1, divNum);
  for ( k = 0; k < divNum; k++ ) {
    sprintf(outFileName, divNumTmpl, OutputBase, k);
    outBaseName = basename(outFileName);
    fprintf(fp, "%s<SNP_File blockPartStart=\"%ld\" blockPartNum=\"%ld\">"
	    "%s</SNP_File>\n", ts2, blockPartStart[k], blockPartNum[k],
	    outBaseName);
  } // end of for(i)
  fprintf(fp, "%s</SNP_Files>\n", ts1);

  fprintf(fp, "</ParaHaplo_DivData>\n");
  fclose(fp);
  return 0;
}
